using System;
using System.Collections.Generic;
using System.Text;
using System.Runtime.InteropServices;

using cc.breeze;

namespace cc.breeze.libpasori
{
    public enum SystemCode : ushort
    {
        ANY = 0xffff,
        SUICA = 0x0003,
        EDY = 0xfe00,
    }

    /// <summary>
    /// libpasori.hのtag_felica構造体
    /// </summary>
    [StructLayout(LayoutKind.Sequential, Pack = 4), ComVisible(false)]
    public sealed class FELICA
    {
        public IntPtr p;
        public SystemCode systemcode;

        [MarshalAs(UnmanagedType.ByValArray, SizeConst = 8)]
        public byte[] IDm;
        [MarshalAs(UnmanagedType.ByValArray, SizeConst = 8)]
        public byte[] PMm;
    }

    /// <summary>
    /// felica型絡み
    /// </summary>
    public class Felica : Observable , IDisposable
    {
        [DllImport("libpasori-dll")]
        static extern IntPtr Felica_polling(IntPtr pp, ushort system, byte RFU, byte timeslot);

        [DllImport("libpasori-dll")]
        static extern void Felica_free(IntPtr fp );

        /// <summary>
        /// 指定されたパソリをポーリングして新しいFelicaオブジェクトを返す。
        /// felica_polling()相当
        /// 今のところ、何も読み取れなかったのかエラーが発生したのか、は
        /// felica_polling()の結果から判断できないので、何も読み取れない場合も例外発生
        /// </summary>
        /// <param name="pp">pasori型へのポインタ</param>
        /// <param name="system">ポーリング対象とするシステムコード。</param>
        /// <param name="RFU"></param>
        /// <param name="timeslot"></param>
        /// <returns></returns>
        public  static Felica Polling( IntPtr pp , SystemCode system , byte RFU , byte timeslot )
        {
            IntPtr fp = Felica_polling(pp, (ushort)system, 0, 0);
            if (fp == IntPtr.Zero)
            {
                throw new LibPasoriException("failed to felica_polling()");
            }

            Felica f = new Felica();
            f.felica_pointer = fp;

            return f;
        }

        /// <summary>
        /// コンストラクタ
        /// </summary>
        private Felica()
        {
        }

        /// <summary>
        /// 取得したfelica*
        /// ここも構造体の形で保持したかったがCインタフェースとのやりとりをIntPtrにすると簡単だったので。
        /// </summary>
        private IntPtr felica_pointer = IntPtr.Zero;
        public IntPtr ToPointer()
        {
            return felica_pointer;
        }

        /// <summary>
        /// felica*から構造体を得る。
        /// </summary>
        /// <returns></returns>
        public FELICA ToStruct()
        {
            return  (FELICA)Marshal.PtrToStructure(this.ToPointer(), typeof(FELICA));
        }

        /// <summary>
        /// felica*から取得したシステムコードを返す
        /// </summary>
        public  SystemCode SystemCode
        {
            get
            {
                return  this.ToStruct().systemcode;
            }
        }

        /// <summary>
        /// felica*から取得したIDmのバイナリ配列を返す。
        /// </summary>
        public byte[] IDm
        {
            get
            {
                return  this.ToStruct().IDm;
            }
        }

        /// <summary>
        /// felica*から取得したPMmのバイナリ配列を返す。
        /// </summary>
        public byte[] PMm
        {
            get
            {
                return  this.ToStruct().PMm;
            }
        }

        /// <summary>
        /// felica*から取得したIDmバイナリ値をHEX文字列に変換して返す。
        /// </summary>
        public string IDm_text
        {
            get
            {
                byte[] b = this.IDm;
                string ret = "";
                for (int i = 0; i < b.Length; i++)
                {
                    ret = ret + b[i].ToString("X2");
                }
                return  ret;
            }
        }

        /// <summary>
        /// felica*から取得したPMmバイナリ値をHEX文字列に変換して返す。
        /// </summary>
        public string PMm_text
        {
            get
            {
                byte[] b = this.PMm;
                string ret = "";
                for (int i = 0; i < b.Length; i++)
                {
                    ret = ret + b[i].ToString("X2");
                }
                return ret;
            }
        }

        /// <summary>
        /// felica*を開放する。
        /// </summary>
        public void Close()
        {
            if (this.ToPointer() != IntPtr.Zero)
            {
                Felica.Felica_free(this.ToPointer());
                this.felica_pointer = IntPtr.Zero;
            }

            this.notifyObserver();
        }

        #region IDisposable メンバ

        public void Dispose()
        {
            this.Close();
        }

        #endregion
    }
}
