using System;
using System.Collections.Generic;
using System.Text;
using System.Runtime.InteropServices;

using cc.breeze;

namespace cc.breeze.libpasori
{

    /// <summary>
    /// libpasori.hのtag_pasori構造体
    /// __LIBPASORI_WITH_LIBUSBが定義されている場合の方。
    /// </summary>
    [StructLayout(LayoutKind.Sequential, Pack = 4), ComVisible(false)]
    public sealed class PASORI
    {
        public IntPtr   dev;
        public IntPtr   dh;
    }

    /// <summary>
    /// pasori型絡み
    /// </summary>
    public class Pasori : IDisposable , IObserver
    {
        [DllImport("libpasori-dll")]
        static extern IntPtr Pasori_open(IntPtr p);

        [DllImport("libpasori-dll")]
        static extern int Pasori_init(IntPtr p);

        [DllImport("libpasori-dll")]
        static extern int Pasori_close(IntPtr p);

        /// <summary>
        /// パソリをオープンして初期化
        /// Pasoriオブジェクトを生成し初期化して返す
        /// </summary>
        /// <returns></returns>
        public static Pasori Open()
        {
            return Pasori.Open(IntPtr.Zero);
        }

        /// <summary>
        /// パソリをオープンして初期化。
        /// Pasoriオブジェクトを生成し初期化して返す
        /// </summary>
        /// <param name="p"></param>
        /// <returns></returns>
        public  static  Pasori Open( IntPtr p )
        {
            Pasori pasori_ = new Pasori();

            IntPtr ps = Pasori_open(p);
            if (ps == IntPtr.Zero)
            {
                throw new LibPasoriException("failed to pasori_open()");
            }

            pasori_.pasori_pointer = ps;
            Pasori_init( pasori_.ToPointer() );

            return pasori_;
        }

        /// <summary>
        /// コンストラクタ
        /// </summary>
        private Pasori()
        {
            this.felica_list = new List<Felica>();
        }

        /// <summary>
        /// パソリをポーリングしてFelicaオブジェクトを返す。
        /// </summary>
        /// <param name="system"></param>
        /// <returns></returns>
        public Felica Polling(SystemCode system)
        {
            return this.Polling(system, 0, 0);
        }

        /// <summary>
        /// パソリをポーリングしてFelicaオブジェクトを返す。
        /// 生成されたFelicaオブジェクトは自身で記録しておく。
        /// </summary>
        /// <param name="system"></param>
        /// <param name="RFU"></param>
        /// <param name="timeslot"></param>
        /// <returns></returns>
        public  Felica  Polling( SystemCode system , byte RFU , byte timeslot )
        {
            Felica fc = Felica.Polling( this.ToPointer() , system , RFU , timeslot );
            this.AddFelica(fc);
            fc.addObserver(this);


            return fc;
        }

        /// <summary>
        /// パソリを閉じてオブジェクト破棄。
        /// Felicaオブジェクトが残っている場合はそれぞれ開放。
        /// libpasori上はPasoriの開放とFelicaの開放をあわせなければならないわけではないが、
        /// felica型にpasori型への参照が含まれるので、Pasori破棄のタイミングでFelicaも破棄することにした。
        /// </summary>
        public  void    Close()
        {
            Felica[] felicas = this.felica_list.ToArray();
            for (int i = 0; i < felicas.Length; i++)
            {
                felicas[i].Close();
            }
            this.felica_list.Clear();

            if (this.ToPointer() != IntPtr.Zero)
            {
                Pasori_close( this.ToPointer() );
                this.pasori_pointer = IntPtr.Zero;
            }
        }

        /// <summary>
        /// pasori型構造体を得る。
        /// </summary>
        /// <returns></returns>
        public PASORI ToStruct()
        {
            return (PASORI)Marshal.PtrToStructure(this.ToPointer(), typeof(PASORI));
        }

        /// <summary>
        /// 取得したpasori型
        /// ホントは構造体のまま保持したかったが、Cインタフェースに対してIntPtrで渡す方が簡単だったから。
        /// </summary>
        private IntPtr pasori_pointer = IntPtr.Zero;

        /// <summary>
        /// pasori型のポインタ
        /// </summary>
        /// <returns></returns>
        public IntPtr ToPointer()
        {
            return pasori_pointer;
        }

        /// <summary>
        /// 生成したFelicaを保持
        /// </summary>
        private List<Felica> felica_list;
        private void AddFelica( Felica f )
        {
            this.felica_list.Add( f );
        }

        /// <summary>
        /// FelicaリストからFelicaを削除する
        /// </summary>
        /// <param name="f"></param>
        private void RemoveFelica(Felica f)
        {
            this.felica_list.Remove(f);
        }

        #region IDisposable メンバ

        public void Dispose()
        {
            this.Close();
        }

        #endregion



        #region IObserver メンバ

        /// <summary>
        /// FelicaがCloseしたら、Pasoriが保持しているFelicaを捨てる。
        /// </summary>
        /// <param name="observable"></param>
        void IObserver.update(Observable observable)
        {
            Felica fc = observable as Felica;
            if (fc != null)
            {
                this.felica_list.Remove(fc);
            }
        }

        #endregion
    }
}



