using System;
using System.Collections.Generic;
using System.Text;
using System.Threading;

namespace cc.breeze.libpasori
{
    /// <summary>
    /// ポーリング完了後コールバック
    /// </summary>
    /// <param name="p"></param>
    /// <param name="f"></param>
    public delegate void PasoriNotifyDelegate(Pasori p, Felica f);

    /// <summary>
    /// ポーリング開始時コールバック
    /// </summary>
    /// <param name="p"></param>
    public delegate void PasoriBeginPollingDelegate(Pasori p);

    /// <summary>
    /// パソリをポーリングし、イベントを通知するクラス。
    /// 以下のタイミングでイベント発生。
    /// ・ポーリング開始時
    /// ・ポーリング完了後
    /// </summary>
    public  class PasoriNotifier
    {
        /// <summary>
        /// コンストラクタ
        /// </summary>
        /// <param name="ps">パソリ</param>
        public PasoriNotifier( Pasori ps )
        {
            this.pasori_ = ps;
        }

        /// <summary>
        /// ポーリング開始
        /// スレッドを起こして、ポーリングループに。
        /// </summary>
        public  void Begin()
        {
            this.polling_thread = new Thread( new ParameterizedThreadStart( PollingProc));
            this.NoMore = false;
            this.polling_thread.Start(this);
        }

        /// <summary>
        /// ポーリング開始
        /// ポーリング完了後のコールバックを指定
        /// </summary>
        /// <param name="cb">ポーリング完了後コールバック</param>
        public void Begin(PasoriNotifyDelegate cb)
        {
            this.NotifyCallback = cb;
            this.Begin();
        }

        /// <summary>
        /// ポーリングループ
        /// NoMoreが真になったらやめる。
        /// </summary>
        /// <param name="o"></param>
        public void PollingProc(object o)
        {
            PasoriNotifyDelegate notify_cb = this.NotifyCallback;
            PasoriBeginPollingDelegate begin_polling_cb = this.BeginPollingCallback;
            Pasori ps = this.pasori_;

            for (; ; )
            {
                if (this.NoMore)
                {
                    break;
                }

                if (begin_polling_callback != null)
                {
                    begin_polling_callback(ps);
                }

                Felica fc = null;
                try
                {
                    fc = ps.Polling(this.SystemCode);
                }
                catch (LibPasoriException ex)
                {
                    /*
                     * エラーがあったのか何も読めなかったのかは区別つかないので握りつぶす
                     */
                }

                /*
                 * エラーの場合もポーリングの終了を知らせねばならないのでCBする
                 */
                notify_cb(ps, fc);

                /*
                 * 寝る。ぐぅ
                 */
                Thread.Sleep(this.Interval);
            }
        }

        /// <summary>
        /// ポーリング開始コールバックを保持
        /// </summary>
        PasoriBeginPollingDelegate begin_polling_callback;
        public PasoriBeginPollingDelegate BeginPollingCallback
        {
            get { return begin_polling_callback; }
            set { begin_polling_callback = value; }
        }

        /// <summary>
        /// ポーリング完了コールバックを保持
        /// </summary>
        PasoriNotifyDelegate notify_callback;
        public PasoriNotifyDelegate NotifyCallback
        {
            get { return notify_callback; }
            set { notify_callback = value; }
        }


        /// <summary>
        /// ポーリングループを終了
        /// スレッドに終了を通知して終了待ち。現在のインターバルの5割増しまで終了を待つ。
        /// 終わらなければ強制終了
        /// </summary>
        public void End()
        {
            this.NoMore = true;
            if (this.polling_thread != null)
            {
                this.polling_thread.Join((this.Interval*3)/2);
                if (this.polling_thread.IsAlive)
                {
                    this.polling_thread.Abort();
                }
            }
        }

        /// <summary>
        /// ポーリング対象システムコードを保持
        /// </summary>
        SystemCode system_code = SystemCode.ANY;
        public SystemCode SystemCode
        {
            get { return system_code; }
            set { system_code = value; }
        }

        /// <summary>
        /// ポーリング終了フラグ
        /// </summary>
        private bool no_more = false;
        private bool NoMore
        {
            get 
            {
                bool ret;
                lock (this.lock_me)
                {
                    ret = this.no_more;
                }
                return ret; 
            }
            set 
            {
                lock (this.lock_me)
                {
                    this.no_more = value;
                }
            }
        }

        object lock_me = new Object();

        /// <summary>
        /// ポーリングスレッド
        /// </summary>
        private Thread polling_thread;

        /// <summary>
        /// ポーリング対象パソリ
        /// </summary>
        private Pasori pasori_;


        /// <summary>
        /// ポーリング間隔。ミリ秒
        /// </summary>
        private int interval_ = 1000;
        /// <summary>
        /// ポーリングを行う間隔。ミリ秒
        /// </summary>
        public int Interval
        {
            get 
            {
                lock (this.lock_me)
                {
                    return interval_;
                }
            }
            set 
            {
                lock (this.lock_me)
                {
                    interval_ = value;
                }
            }
        }
    }
}
